#
# (C) Copyright 2014 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

from __future__ import absolute_import

from .stream import BigEndianStream, LittleEndianStream
from .constants import TYPE_NAMES, BITFIELD, MAGIC, ENDIAN_MARKER, FORMAT_VERSION_NUMBER_MAJOR, FORMAT_VERSION_NUMBER_MINOR, NEW_HEADER
from .codec import read_codec

from collections import Iterable
from itertools import accumulate, chain
import pandas as pd


class MismatchedFramesError(ValueError):
    pass


class ColumnInfo:
    """
    Represent the type of a column
    """

    class Bitfield:
        def __init__(self, name, size, offset):
            self.name = name
            self.size = size
            self.offset = offset

        def __eq__(self, other):
            return (self.name == other.name and
                    self.size == other.size and
                    self.offset == other.offset)

    def __init__(self, name, idx, dtype, datasize, bitfields):
        self.name = name
        self.dtype = dtype
        self.index = idx
        self.datasize = datasize
        self.bitfields = bitfields
        assert (dtype == BITFIELD) != (len(bitfields) == 0)
        if self.bitfields:
            assert isinstance(self.bitfields, Iterable)
            assert all(isinstance(b, ColumnInfo.Bitfield) for b in self.bitfields)

    def __str__(self):
        if self.dtype == BITFIELD:
            bitfield_str = "(" + ",".join("{}:{}".format(b.name, b.size) for b in self.bitfields) + ")"
        else:
            bitfield_str = ""
        return "{}:{}{}".format(self.name, TYPE_NAMES.get(self.dtype, '<unknown>'), bitfield_str)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (self.name == other.name and
                self.dtype == other.dtype and
                self.index == other.index and    # This may be overzealous?
                self.datasize == other.datasize and
                self.bitfields == other.bitfields)



class Frame:

    def __init__(self, f):

        # Read marker and magic

        m = f.read(2)
        if len(m) == 0:
            raise EOFError()
        assert int.from_bytes(m, byteorder='big', signed=False) == NEW_HEADER

        assert f.read(3) == MAGIC

        # Get byte ordering

        endian_marker = f.read(4)
        if int.from_bytes(endian_marker, byteorder='little') == ENDIAN_MARKER:
            stream = LittleEndianStream(f)
        else:
            stream = BigEndianStream(f)

        # TODO: Some inequalities here, rather than plain statement
        assert stream.readInt32() == FORMAT_VERSION_NUMBER_MAJOR
        assert stream.readInt32() == FORMAT_VERSION_NUMBER_MINOR

        md5 = stream.readString()

        headerLength = stream.readInt32()
        self._dataStartPosition = stream.position() + headerLength

        # 0 means we don't know offset of next header.

        self._dataSize = stream.readInt64()
        self._dataEndPosition = self._dataStartPosition + self._dataSize

        # prevFrameOffset
        assert stream.readInt64() == 0

        self._numberOfRows = stream.readInt64()

        self.flags = [stream.readReal64() for _ in range(stream.readInt32())]
        self.properties = {stream.readString(): stream.readString() for _ in range(stream.readInt32())}

        self._numberOfColumns = stream.readInt32()
        self._columnPosition = stream.position()

        self.__columnCodecs = None
        self._columns = None

        self._stream = stream

        # Support frame aggregation
        self._trailingAggregatedFrames = []

    def seekToEnd(self):
        self._stream.seek(self._dataEndPosition)

    @property
    def _column_codecs(self):
        """
        Internal method to get the codecs for the given column.
        These are read/constructed lazily from the file handle so that we can do scans through the file rapdily.

        :return: A list of coders
        """
        if self.__columnCodecs is None:
            self._stream.seek(self._columnPosition)
            self.__columnCodecs = [read_codec(self._stream) for _ in range(self._numberOfColumns)]
            assert self._stream.position() == self._dataStartPosition
        return self.__columnCodecs

    @property
    def columns(self):
        return [
            ColumnInfo(codec.column_name, idx, codec.type, codec.data_size,
                       [ColumnInfo.Bitfield(name=nm, size=sz, offset=off)
                        for nm, sz, off in zip(codec.bitfield_names,
                                               codec.bitfield_sizes,
                                               accumulate(chain([0]), codec.bitfield_sizes))])
            for idx, codec in enumerate(self._column_codecs)
        ]

    @property
    def column_dict(self):
        return {c.name: c for c in self.columns}

    @property
    def simple_column_dict(self):
        return {c.name.split('@')[0]: c for c in self.columns}

    @property
    def nrows(self):
        return self._numberOfRows + sum(f.nrows for f in self._trailingAggregatedFrames)

    @property
    def ncolumns(self):
        if self.__columnCodecs is not None:
            assert self._numberOfColumns == len(self.__columnCodecs)
        return self._numberOfColumns

    def dataframe(self, columns=None):
        """
        Actually decode the data!
        TODO: Properly skip decoding columns that aren't needed
        """
        column_codecs = self._column_codecs

        self._stream.seek(self._dataStartPosition)

        output_cols = [[] for _ in range(self._numberOfColumns)]
        output = {c.column_name: data for c, data in zip(column_codecs, output_cols)
                  if columns is None or c.column_name in columns}
        lastDecoded = [0] * self._numberOfColumns

        lastStartCol = 0
        for row in range(self._numberOfRows):

            startCol = self._stream.readMarker()

            # Fill in missing data in columns
            if lastStartCol > startCol:
                for col in range(startCol, lastStartCol):
                    last = output_cols[col][-1]
                    output_cols[col].extend(last for _ in range(row - lastDecoded[col] - 1))

            lastStartCol = startCol

            for col in range(startCol, self._numberOfColumns):
                output_cols[col].append(column_codecs[col].decode(self._stream))
                lastDecoded[col] = row

        for col in range(lastStartCol):
            last = output_cols[col][-1]
            output_cols[col].extend(last for _ in range(self._numberOfRows - lastDecoded[col] - 1))

        df = pd.DataFrame(output)

        if len(self._trailingAggregatedFrames) > 0:
            return pd.concat([df] + [f.dataframe(columns) for f in self._trailingAggregatedFrames], copy=False, axis=0)
        else:
            return df

    def _append(self, frame: 'Frame'):
        if self.column_dict != frame.column_dict:
            raise MismatchedFramesError
        self._trailingAggregatedFrames.append(frame)





from .constants import *
from .lib import lib, ffi
import pandas

def encode_odb(df: pandas.DataFrame, f, types: dict=None, rows_per_frame=10000, properties=None, **kwargs):
    """
    Encode a pandas dataframe into ODB2 format

    :param df: The dataframe to encode
    :param f: The file-like object into which to encode the ODB2 data
    :param types: An optional (sparse) dictionary. Each key-value pair maps the name of a column to
                  encode to an ODB2 data type to use to encode it.
    :param rows_per_frame: The maximum number of rows to encode per frame. If this number is exceeded,
                           a sequence of frames will be encoded
    :param kwargs: Accept extra arguments that may be used by the python pyodc encoder.
    :return:
    """
    if isinstance(f, str):
        with open(f, 'wb') as freal:
            return encode_odb(df, freal, types=types, rows_per_frame=rows_per_frame, **kwargs)

    # Some constants that are useful

    pmissing_integer = ffi.new('long*')
    pmissing_double = ffi.new('double*')
    lib.odc_missing_integer(pmissing_integer)
    lib.odc_missing_double(pmissing_double)
    missing_integer = pmissing_integer[0]
    missing_double = pmissing_double[0]

    def infer_column_type(arr, override_type):
        """
        Given a column of data, infer the encoding type.
        :param arr: The column of data to encode
        :param override_type:
        :return: (return_arr, dtype)
            - return_arr is the column of data to encode. This may be of a different internal type/contents
              to that supplied to the function, but it will normally not be.
            - The ODB2 type to encode with.
        """
        return_arr = arr
        dtype = override_type

        if dtype is None:
            if arr.dtype in ('uint64', 'int64'):
                dtype = INTEGER
            elif arr.dtype == 'float64':
                if not data.isnull().all() and all(pandas.isnull(v) or float(v).is_integer() for v in arr):
                    dtype = INTEGER
                    return_arr = arr.fillna(value=missing_integer).astype('int64')
                else:
                    dtype = DOUBLE
                    return_arr = arr.fillna(value=missing_double)
            elif arr.dtype == 'object':
                if not arr.isnull().all() and all(s is None or isinstance(s, str) for s in arr):
                    dtype = STRING
                elif arr.isnull().all():
                    dtype = INTEGER

        if arr.dtype == 'object':
            # Map strings into an array that can be read in C
            if dtype == STRING:
                return_arr = return_arr.astype("|S{}".format(max(8, 8 * (1 + ((max(len(s) for s in arr)- 1) // 8)))))
            elif dtype == INTEGER:
                return_arr = return_arr.fillna(value=missing_integer).astype("int64")

        if dtype is None:
            raise ValueError("Unsupported value type: {}".format(arr.dtype))

        return return_arr, dtype

    nrows = df.shape[0]
    if types is None:
        types = {}

    encoder = ffi.new('odc_encoder_t**')
    lib.odc_new_encoder(encoder)
    encoder = ffi.gc(encoder[0], lib.odc_free_encoder)

    for k, v in (properties or {}).items():
        lib.odc_encoder_add_property(encoder, k.encode('utf-8'), v.encode('utf-8'))

    lib.odc_encoder_set_row_count(encoder, nrows)
    lib.odc_encoder_set_rows_per_frame(encoder, rows_per_frame)

    # We store all of the numpy arrays here. Mostly this is just another reference to an
    # existing array, but some of the types require us to create a new (casted) copy, so
    # we need to put it somewhere to ensure it stays alive appropriately long.
    data_cache = []

    for i, (name, data) in enumerate(df.items()):
        data, dtype = infer_column_type(data, types.get(name, None))
        data_cache.append(data)

        lib.odc_encoder_add_column(encoder, name.encode('utf-8'), dtype)
        lib.odc_encoder_column_set_data_array(encoder, i, data.dtype.itemsize, data.array.to_numpy().strides[0],
                                              ffi.cast('void*', data.values.ctypes.data))

    lib.odc_encode_to_file_descriptor(encoder, f.fileno(), ffi.NULL)




from .constants import MISSING_INTEGER, DataType, MISSING_REAL, INTERNAL_REAL_MISSING, MISSING_STRING

import struct
import pandas as pd


class Codec:

    def __init__(self, column_name: str, minval: float, maxval: float, data_type: DataType, has_missing: bool=False,
                 bitfield_names=None, bitfield_sizes=None):
        self.column_name = column_name
        self.min = minval
        self.max = maxval
        self.has_missing = has_missing
        self.bitfield_names = bitfield_names
        self.bitfield_sizes = bitfield_sizes

        self.type = data_type
        self.missing_value = {
            DataType.INTEGER: MISSING_INTEGER,
            DataType.BITFIELD: MISSING_INTEGER,
            DataType.DOUBLE: MISSING_REAL,
            DataType.REAL: MISSING_REAL,
            DataType.STRING: MISSING_STRING
        }[data_type]

        assert self.name is not None

    @property
    def name(self):
        return ''.join('_' + c.lower() if c.isupper() else c for c in self.__class__.__name__).strip('_')

    @staticmethod
    def codec_class_name(name):
        return ''.join(part[:1].upper() + part[1:] for part in name.split("_"))

    @property
    def data_size(self):
        """
        Size of the decoded data in bytes.
        """
        return 8

    def encode_header(self, stream):
        stream.encodeString(self.column_name)
        stream.encodeInt32(self.type)
        # TODO: Bitfield stuff goes here
        stream.encodeString(self.name)
        stream.encodeInt32(1 if self.has_missing else 0)
        stream.encodeReal64(self.min)
        stream.encodeReal64(self.max)
        stream.encodeReal64(self.missing_value)

    @staticmethod
    def read_core_header(stream):
        has_missing = False if stream.readInt32() == 0 else True
        minval = stream.readReal64()
        maxval = stream.readReal64()
        missing_value = stream.readReal64()
        return has_missing, minval, maxval, missing_value

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):
        raise NotImplementedError

    @classmethod
    def from_stream(cls, stream, column_name: str, data_type: DataType, bitfield_names, bitfield_sizes):
        has_missing, minval, maxval, missing_value = cls.read_core_header(stream)
        return cls(column_name, minval, maxval, data_type, has_missing=has_missing,
                   bitfield_names=bitfield_names, bitfield_sizes=bitfield_sizes)

    def encode(self, stream, value):
        raise NotImplementedError

    def decode(self, stream):
        raise NotImplementedError

    @property
    def numChanges(self):
        raise NotImplementedError


# n.b. The codec names match the class names

class Constant(Codec):

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):
        assert data.nunique() == 1 and not data.hasnans
        value = next(iter(data))
        return cls(column_name, value, value, data_type)

    def encode(self, stream, value):
        pass

    def decode(self, stream):
        return {
            DataType.INTEGER: int,
            DataType.REAL: float,
            DataType.DOUBLE: float,
            DataType.BITFIELD: int
        }[self.type](self.min)

    @property
    def numChanges(self):
        return 0


class ConstantString(Constant):

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):
        assert data.nunique() == 1 and not data.hasnans
        assert data_type == DataType.STRING

        # n.b. This looks like it ties it to little-endian, but it doesn't. Byte order
        #      is always the same for string data, but we are 'pretending' to be a double.
        value = struct.unpack('<d', (next(iter(data)).encode('utf-8') + (b'\x00' * 8))[:8])[0]

        return cls(column_name, value, value, data_type)

    def decode(self, stream):
        return struct.pack('<d', self.min).decode('utf-8').strip('\x00')


class NumericBase(Codec):

    _numChanges = None
    _data = None
    accepted_types = None

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):
        assert data_type in cls.accepted_types
        c = cls(column_name, data.min(), data.max(), data_type, has_missing=data.hasnans)
        c._data = data
        return c

    @property
    def numChanges(self):
        if self._numChanges is None:
            assert self._data is not None
            self._numChanges = (self._data.fillna(self.missing_value).diff() != 0).sum() - 1
        return self._numChanges


class ConstantOrMissing(NumericBase):

    internal_missing_value = 0xff
    accepted_types = (DataType.INTEGER, DataType.BITFIELD)

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):
        assert data.nunique() == 1 and data.hasnans
        return super().from_dataframe(column_name, data, data_type)

    def encode(self, stream, value):
        if pd.isnull(value):
            stream.encodeUInt8(self.internal_missing_value)
        else:
            stream.encodeUInt8(0)

    def decode(self, stream):
        marker = stream.readUInt8()
        if marker == self.internal_missing_value:
            return None
        else:
            return {
                DataType.INTEGER: int,
                DataType.REAL: float,
                DataType.DOUBLE: float,
                DataType.BITFIELD: int
            }[self.type](self.min)


class RealConstantOrMissing(ConstantOrMissing):

    accepted_types = (DataType.DOUBLE, DataType.REAL)


class OffsetInteger(NumericBase):

    internal_missing_value = None
    max_range = 0
    accepted_types = (DataType.INTEGER, DataType.BITFIELD)

    @staticmethod
    def _encode(stream, value):
        raise NotImplementedError

    @staticmethod
    def _decode(stream):
        raise NotImplementedError

    def encode(self, stream, value):
        if pd.isnull(value):
            self._encode(stream, self.internal_missing_value)
        else:
            assert float(value).is_integer()
            assert value - self.min <= self.max_range
            self._encode(stream, int(value - self.min))

    def decode(self, stream):
        value = self._decode(stream)
        return None if value == self.internal_missing_value else int(value + self.min)


class Int8(OffsetInteger):

    max_range = 0xff
    accepts_missing = False

    @staticmethod
    def _encode(stream, value):
        stream.encodeUInt8(value)

    @staticmethod
    def _decode(stream):
        return stream.readUInt8()


class Int8Missing(Int8):
    max_range = 0xfe
    accepts_missing = True
    internal_missing_value = 0xff


class Int16(OffsetInteger):

    max_range = 0xffff
    accepts_missing = False

    @staticmethod
    def _encode(stream, value):
        stream.encodeUInt16(value)

    @staticmethod
    def _decode(stream):
        return stream.readUInt16()


class Int16Missing(Int16):
    max_range = 0xfffe
    accepts_missing = True
    internal_missing_value = 0xffff


class Int32(NumericBase):

    max_range = 0xfffffffe
    accepts_missing = True
    internal_missing_value = 0x7fffffff
    accepted_types = (DataType.INTEGER, DataType.BITFIELD)

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):
        if data.min() < -0x80000000 or data.max() >= 0x7fffffff:
            raise ValueError("Cannot encode integers out of range")
        c = super().from_dataframe(column_name, data, data_type)
        assert c.missing_value == c.internal_missing_value
        return c

    def encode(self, stream, value):
        if pd.isnull(value):
            stream.encodeInt32(self.internal_missing_value)
        else:
            stream.encodeInt32(int(value))

    def decode(self, stream):
        value = stream.readInt32()
        return None if value == self.internal_missing_value else value


class LongReal(NumericBase):

    accepted_types = (DataType.DOUBLE, DataType.REAL)

    def encode(self, stream, value):
        stream.encodeReal64(value)

    def decode(self, stream):
        return stream.readReal64()


class ShortReal(NumericBase):

    internal_missing_value = INTERNAL_REAL_MISSING[0]
    accepted_types = (DataType.DOUBLE, DataType.REAL)

    def encode(self, stream, value):
        if pd.isnull(value) is None:
            stream.encodeReal32(self.internal_missing_value)
        else:
            assert value != self.internal_missing_value
            stream.encodeReal32(value)

    def decode(self, stream):
        value = stream.readReal32()
        return None if value == self.internal_missing_value else value


class ShortReal2(ShortReal):
    internal_missing_value = INTERNAL_REAL_MISSING[1]


class Int8String(Codec):

    missing_value = MISSING_INTEGER
    type = DataType.STRING
    _numChanges = None

    def __init__(self, *args, values=None, data=None, **kwargs):

        self._data = data
        assert values is not None
        self.values = values
        self.value_map = {value: i for i, value in enumerate(values)}

        super().__init__(*args, **kwargs)

    @classmethod
    def from_dataframe(cls, column_name: str, data: pd.Series, data_type: DataType):

        assert not data.hasnans
        assert data_type == DataType.STRING
        return cls(column_name, 0, 0, data_type, values=data.unique(), data=data)

    @classmethod
    def from_stream(cls, stream, column_name: str, data_type: DataType, bitfield_names, bitfield_sizes):
        has_missing, minval, maxval, missing_value = cls.read_core_header(stream)
        nvalues = stream.readInt32()
        values = [None] * nvalues
        for _ in range(nvalues):
            value = stream.readString()
            stream.readInt32()  # discard
            idx = stream.readInt32()
            values[idx] = value
        assert not any(v is None for v in values)
        return cls(column_name, minval, maxval, data_type, values=values, has_missing=has_missing,
                   bitfield_names=bitfield_names, bitfield_sizes=bitfield_sizes)

    @staticmethod
    def _encode(stream, value):
        stream.encodeUInt8(value)

    @staticmethod
    def _decode(stream):
        return stream.readUInt8()

    def encode_header(self, stream):
        super().encode_header(stream)
        stream.encodeInt32(len(self.value_map))
        for value, idx in self.value_map.items():
            stream.encodeString(value)
            stream.encodeInt32(0)
            stream.encodeInt32(idx)

    def encode(self, stream, value):
        idx = self.value_map[value]
        self._encode(stream, idx)

    def decode(self, stream):
        idx = self._decode(stream)
        return self.values[idx]

    @property
    def numChanges(self):
        if self._numChanges is None:
            assert self._data is not None
            self._numChanges = (self._data != self._data.shift()).sum() - 1
        return self._numChanges


class Int16String(Int8String):

    @staticmethod
    def _encode(stream, value):
        stream.encodeUInt16(value)

    @staticmethod
    def _decode(stream):
        return stream.readUInt16()


def select_codec(column_name: str, data: pd.Series, data_type):

    # If data types are not specified, determine them from the pandas Series

    if data_type is None:
        if data.dtype in ["{}int{}".format(s, b) for s in ('', 'u') for b in (8, 16, 32, 64)]:
            data_type = DataType.INTEGER
        elif data.dtype == 'float64':
            if not data.isnull().all() and all(pd.isnull(v) or float(v).is_integer() for v in data):
                data_type = DataType.INTEGER
            else:
                data_type = DataType.DOUBLE
        elif data.dtype == 'float32':
            data_type = DataType.REAL
        elif data.dtype == 'object':
            if not data.isnull().all() and all(s is None or isinstance(s, str) for s in data):
                data_type = DataType.STRING

    if data_type is None:
        raise RuntimeError("Unknown data type {} not supported".format(data.dtype))

    # And now the logic for selecting the codec

    codec_class = None

    if data.nunique() == 1 and not data.hasnans:

        codec_class = Constant

    if data_type == DataType.INTEGER:

        range = data.max() - data.min()
        has_missing = data.hasnans

        for c in [Int8, Int8Missing, Int16, Int16Missing, Int32]:
            if range <= c.max_range and (c.accepts_missing or not has_missing):
                codec_class = c
                break

    elif data_type == DataType.DOUBLE:

        codec_class = LongReal

    elif data_type == DataType.REAL:

        if INTERNAL_REAL_MISSING[1] in data:
            if INTERNAL_REAL_MISSING[0] in data:
                raise ValueError("Cannot encode a float data series with both internal missing values")
            codec_class = ShortReal
        else:
            codec_class = ShortReal2

    elif data_type == DataType.STRING:

        if data.nunique() <= 256:
            codec_class = Int8String
        else:
            assert data.nunique() <= 32767
            codec_class = Int16String

    if codec_class is not None:
        return codec_class.from_dataframe(column_name, data, data_type)

    print(data)
    print(data_type)
    raise NotImplementedError


def read_codec(stream):
    column_name = stream.readString()
    data_type = DataType(stream.readInt32())

    if data_type == DataType.BITFIELD:
        bitFieldNames = [stream.readString() for _ in range(stream.readInt32())]
        bitFieldSizes = [stream.readInt32() for _ in range(stream.readInt32())]
    else:
        bitFieldNames = []
        bitFieldSizes = []

    codec_class = stream.readString()
    codec_class = globals()[Codec.codec_class_name(codec_class)]
    return codec_class.from_stream(stream, column_name, data_type, bitFieldNames, bitFieldSizes)





from .lib import lib
from enum import IntEnum, unique


@unique
class DataType(IntEnum):
    IGNORE = lib.ODC_IGNORE
    INTEGER = lib.ODC_INTEGER
    DOUBLE = lib.ODC_DOUBLE
    REAL = lib.ODC_REAL
    STRING = lib.ODC_STRING
    BITFIELD = lib.ODC_BITFIELD


IGNORE = DataType.IGNORE
INTEGER = DataType.INTEGER
REAL = DataType.REAL
STRING = DataType.STRING
BITFIELD = DataType.BITFIELD
DOUBLE = DataType.DOUBLE




from enum import IntEnum, unique
import struct

# Current version information

FORMAT_VERSION_NUMBER_MAJOR = 0
FORMAT_VERSION_NUMBER_MINOR = 5

# Default missing values

MISSING_INTEGER = 2147483647
MISSING_REAL = -2147483647
MISSING_STRING = struct.unpack('<d', b'\x00\x00\x00\x00\x00\x00\x00\x00')[0]

INTERNAL_REAL_MISSING = (
    struct.unpack("<f", b"\x00\x00\x80\x00")[0],
    struct.unpack("<f", b"\xff\xff\x7f\x7f")[0]
)

# Constants for header encoding

NEW_HEADER = 65535
MAGIC = b"ODA"
ENDIAN_MARKER = 1

# Types

@unique
class DataType(IntEnum):

    IGNORE = 0
    INTEGER = 1
    REAL = 2
    STRING = 3
    BITFIELD = 4
    DOUBLE = 5

IGNORE = DataType.IGNORE
INTEGER = DataType.INTEGER
REAL = DataType.REAL
STRING = DataType.STRING
BITFIELD = DataType.BITFIELD
DOUBLE = DataType.DOUBLE

TYPE_NAMES = {
    IGNORE: 'ignore',
    INTEGER: 'integer',
    REAL: 'real',
    STRING: 'string',
    BITFIELD: 'bitfield',
    DOUBLE: 'double'
}



#
# Copyright 2017-2020 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pkg_resources import parse_version
import cffi
import os

__odc_version__ = "1.0.2"

ffi = cffi.FFI()


class ODCException(RuntimeError):
    pass


class CFFIModuleLoadFailed(ODCException):
    pass


class PatchedLib:
    """
    Patch a CFFI library with error handling

    Finds the header file associated with the ODC C API and parses it, loads the shared library,
    and patches the accessors with automatic python-C error handling.
    """
    __type_names = {}

    def __init__(self):

        ffi.cdef(self.__read_header())
        try:
            self.__lib = ffi.dlopen('odccore')
        except Exception as e:
            raise CFFIModuleLoadFailed() from e

        # Todo: Version check against __version__

        # All of the executable members of the CFFI-loaded library are functions in the ODC
        # C API. These should be wrapped with the correct error handling. Otherwise forward
        # these on directly.

        for f in dir(self.__lib):
            try:
                attr = getattr(self.__lib, f)
                setattr(self, f, self.__check_error(attr, f) if callable(attr) else attr)
            except Exception as e:
                print(e)
                print("Error retrieving attribute", f, "from library")

        # Initialise the library, and sett it up for python-appropriate behaviour

        self.odc_initialise_api()
        self.odc_integer_behaviour(self.ODC_INTEGERS_AS_LONGS)

        # Check the library version

        tmp_str = ffi.new('char**')
        self.odc_version(tmp_str)
        versionstr = ffi.string(tmp_str[0]).decode('utf-8')

        if parse_version(versionstr) < parse_version(__odc_version__):
            raise RuntimeError("Version of libodc found is too old. {} < {}".format(versionstr, __odc_version__))

    def type_name(self, dtype: 'DataType'):
        name = self.__type_names.get(dtype, None)
        if name is not None:
            return name

        name_tmp = ffi.new('char**')
        self.odc_column_type_name(dtype, name_tmp)
        name = ffi.string(name_tmp[0]).decode('utf-8')
        self.__type_names[dtype] = name
        return name

    def __read_header(self):
        with open(os.path.join(os.path.dirname(__file__), 'processed_odc.h'), 'r') as f:
            return f.read()

    def __check_error(self, fn, name):
        """
        If calls into the ODC library return errors, ensure that they get detected and reported
        by throwing an appropriate python exception.
        """
        def wrapped_fn(*args, **kwargs):
            retval = fn(*args, **kwargs)
            if retval not in (self.__lib.ODC_SUCCESS, self.__lib.ODC_ITERATION_COMPLETE):
                error_str = "Error in function {}: {}".format(name, self.__lib.odc_error_string(retval))
                raise ODCException(error_str)
            return retval

        return wrapped_fn


def memoize_constant(fn):
    """Memoize constant values to avoid repeatedly crossing the API layer unecessarily"""
    attr_name = "__memoized_{}".format(fn.__name__)

    def wrapped_fn(self):
        value = getattr(self, attr_name, None)
        if value is None:
            value = fn(self)
            setattr(self, attr_name, value)
        return value

    return wrapped_fn


# Bootstrap the library

try:
    lib = PatchedLib()
except CFFIModuleLoadFailed as e:
    raise ImportError() from e




from .constants import (NEW_HEADER, MAGIC, FORMAT_VERSION_NUMBER_MAJOR, FORMAT_VERSION_NUMBER_MINOR,
                        ENDIAN_MARKER)
from .codec import select_codec
from .stream import LittleEndianStream, BigEndianStream

import hashlib
import pandas as pd
import numpy as np
import io


def encode_odb(dataframe: pd.DataFrame, f, rows_per_frame=10000, types=None, bigendian: bool=False,
               properties: dict=None):
    """
    :param dataframe: A pandas dataframe to encode
    :param f: A file-like object to write the encoded data to
    :param columns: A dict of (optional) column-name : constant DataType pairs, or None
    :param bigendian: Encode in big- endian byte order if True
    :param properties: Encode a dictionary of supplied properties
    """
    if isinstance(f, str):
        with open(f, 'wb') as freal:
            return encode_odb(dataframe, freal, rows_per_frame=rows_per_frame, types=types, bigendian=bigendian)

    column_order = None

    # Split the dataframe into chunks of appropriate size
    for i, sub_df in dataframe.groupby(np.arange(len(dataframe)) // rows_per_frame):
        column_order = encode_single_dataframe(sub_df, f,
                                               types=types,
                                               column_order=column_order,
                                               bigendian=bigendian,
                                               properties=(properties or {}))


def encode_single_dataframe(dataframe: pd.DataFrame, f, types: dict=None, column_order: list=None,
                            bigendian: bool=False, properties: dict=None):
    """
    :param dataframe: A pandas dataframe to encode
    :param f: A file-like object to write the encoded data to
    :param columns: A dict of (optional) column-name : constant DataType pairs, or None
    :param column_order: A list of column names specifying the encode order. If None, optimise according
                         to the rate of value changes in the columns
    :param bigendian: Encode in big- endian byte order if True
    :param properties: Encode a dictionary of supplied properties
    :return: The column order used for encoding as a list of column names
    """

    stream_class = BigEndianStream if bigendian else LittleEndianStream

    codecs = [select_codec(name, data, (types or {}).get(name, None)) for name, data in dataframe.iteritems()]

    # If a column order has been specified, sort the codecs according to it. otherwise sort
    # the codecs for the most efficient use of the given data

    if column_order:
        assert len(column_order) == len(set(column_order))
        assert set(column_order) == set(c.column_name for c in codecs)
        codecs = {c.column_name: c for c in codecs}
        codecs = [codecs[column_name] for column_name in column_order]
    else:
        codecs.sort(key=lambda c: c.numChanges)
        column_order = [c.column_name for c in codecs]
        assert len(column_order) == len(set(column_order))

    data = _encodeData(dataframe, codecs, stream_class)
    headerPart2 = _encodeHeaderPart2(dataframe, codecs, stream_class, len(data), (properties or {}))
    headerPart1 = _encodeHeaderPart1(headerPart2, stream_class)

    f.write(headerPart1)
    f.write(headerPart2)
    f.write(data)

    return column_order


def _encodeData(dataframe, codecs, stream_class):
    """
    Encode the data into a memory buffer
    """
    dataIO = io.BytesIO()
    stream = stream_class(dataIO)

    # Encode the column in the order supplied in the indexes list, rather than that
    # inherent in the dataframe.

    column_indexes = {column_name: i for i, column_name in enumerate(dataframe.columns)}
    column_indexes = [column_indexes[codec.column_name] for codec in codecs]

    # Iterate over rows

    last_row = None
    codec_indexes = list(zip(codecs, column_indexes))

    for row in dataframe.itertuples(index=False):

        for i, (codec, index) in enumerate(codec_indexes):
            if last_row is None or (row[index] != last_row[index] and not (pd.isnull(row[index]) and pd.isnull(last_row[index]))):
                break

        stream.encodeMarker(i)

        for codec, index in codec_indexes[i:]:
            codec.encode(stream, row[index])

        last_row = row

    return dataIO.getbuffer()


def _encodeHeaderPart2(dataframe, codecs, stream_class, data_len, properties):
    """
    Encode the second part of the header - whose size and md5 is required to encode
    the first part of the header.
    """
    headerIO = io.BytesIO()
    stream = stream_class(headerIO)

    # Data Size (nextFrameOffset)
    stream.encodeInt64(data_len)

    # prevFrameOffset = 0
    stream.encodeInt64(0)

    # Number of rows
    stream.encodeInt64(dataframe.shape[0])

    # Flags... --> Currently no flags, no properties (remember to update Header Length)
    stream.encodeInt32(0)

    # Externally specified properties
    stream.encodeInt32(len(properties))
    for k, v in properties.items():
        stream.encodeString(k)
        stream.encodeString(v)

    # Number of columns
    stream.encodeInt32(dataframe.shape[1])

    # Encode the codec information
    for codec in codecs:
        codec.encode_header(stream)

    return headerIO.getbuffer()


def _encodeHeaderPart1(headerPart2, stream_class):
    """
    Encode the first part of the header. This _could_ be done directly
    on the file-like object, although that would incurr many more filesystem
    operations.
    """
    headerIO = io.BytesIO()
    stream = stream_class(headerIO)

    stream.encodeMarker(NEW_HEADER)
    stream.write(MAGIC)

    stream.encodeInt32(ENDIAN_MARKER)

    stream.encodeInt32(FORMAT_VERSION_NUMBER_MAJOR)
    stream.encodeInt32(FORMAT_VERSION_NUMBER_MINOR)

    # MD5
    m = hashlib.md5()
    m.update(headerPart2)
    md = m.hexdigest()
    stream.encodeString(md)

    # Header Length
    stream.encodeInt32(len(headerPart2))

    return headerIO.getbuffer()







from .constants import IGNORE, INTEGER, REAL, STRING, BITFIELD, DOUBLE, DataType
from .frame import Frame, ColumnInfo
from .encoder import encode_odb
from .reader import Reader, read_odb
from .lib import ODCException

__version__ = '1.0.2'



from .constants import *
from .lib import ffi, lib, memoize_constant

from collections import Iterable
import numpy
import pandas
import os


class ColumnInfo:

    class Bitfield:
        def __init__(self, name, size, offset):
            self.name = name
            self.size = size
            self.offset = offset

    def __init__(self, name, idx, dtype, datasize, bitfields):
        self.name = name
        self.dtype = dtype
        self.index = idx
        self.datasize = datasize
        self.bitfields = bitfields
        assert (dtype == BITFIELD) != (bitfields is None)
        if self.bitfields:
            assert isinstance(self.bitfields, Iterable)
            assert all(isinstance(b, ColumnInfo.Bitfield) for b in self.bitfields)

    def __str__(self):
        if self.bitfields is not None:
            bitfield_str = "(" + ",".join("{}:{}".format(b[0], b[1]) for b in self.bitfields) + ")"
        else:
            bitfield_str = ""
        return "{}:{}{}".format(self.name, self.dtype, bitfield_str)

    def __repr__(self):
        return str(self)


class Frame:

    def __init__(self, table):
        self.__frame = table
        self.__columns = None

    @property
    @memoize_constant
    def columns(self):
        columns = []
        for col in range(self.ncolumns):

            pname = ffi.new('const char**')
            ptype = ffi.new('int*')
            pdatasize = ffi.new('int*')
            pbitfield_count = ffi.new('int*')
            lib.odc_frame_column_attributes(self.__frame, col, pname, ptype, pdatasize, pbitfield_count)
            name = ffi.string(pname[0]).decode('utf-8')
            dtype = DataType(int(ptype[0]))
            datasize = int(pdatasize[0])
            bitfield_count = int(pbitfield_count[0])
            bitfields = None

            if dtype == STRING:
                assert datasize % 8 == 0
            else:
                assert datasize == 8

            if dtype == BITFIELD:
                bitfields = []
                for n in range(bitfield_count):

                    pbitfield_name = ffi.new('const char**')
                    poffset = ffi.new('int*')
                    psize = ffi.new('int*')
                    lib.odc_frame_bitfield_attributes(self.__frame, col, n, pbitfield_name, poffset, psize)

                    bitfields.append(ColumnInfo.Bitfield(
                        name=ffi.string(pbitfield_name[0]).decode('utf-8'),
                        size=int(psize[0]),
                        offset=int(poffset[0])))

            columns.append(ColumnInfo(name, col, dtype, datasize, bitfields))

        return columns

    @property
    @memoize_constant
    def column_dict(self):
        return {c.name: c for c in self.columns}

    @property
    @memoize_constant
    def simple_column_dict(self):
        return {c.name.split('@')[0]: c for c in self.columns}

    @property
    @memoize_constant
    def nrows(self):
        count = ffi.new('long*')
        lib.odc_frame_row_count(self.__frame, count)
        return int(count[0])

    @property
    @memoize_constant
    def ncolumns(self):
        count = ffi.new('int*')
        lib.odc_frame_column_count(self.__frame, count)
        return int(count[0])

    def dataframe(self, columns=None):

        # Some constants that are useful

        pmissing_integer = ffi.new('long*')
        pmissing_double = ffi.new('double*')
        lib.odc_missing_integer(pmissing_integer)
        lib.odc_missing_double(pmissing_double)
        missing_integer = pmissing_integer[0]
        missing_double = pmissing_double[0]

        if columns is None:
            columns = [c.name for c in self.columns]

        assert columns is not None

        cd = self.column_dict
        scd = self.simple_column_dict

        integer_cols = []
        double_cols = []
        string_cols = {}
        for name in columns:
            try:
                col = cd[name]
            except KeyError:
                col = scd[name]
            if col.dtype == INTEGER or col.dtype == BITFIELD:
                integer_cols.append(col)
            elif col.dtype == REAL or col.dtype == DOUBLE:
                double_cols.append(col)
            elif col.dtype == STRING:
                string_cols.setdefault(col.datasize, []).append(col)

        decoder = ffi.new('odc_decoder_t**')
        lib.odc_new_decoder(decoder)
        decoder = ffi.gc(decoder[0], lib.odc_free_decoder)

        lib.odc_decoder_set_row_count(decoder, self.nrows)

        dataframes = []
        pos = 0
        string_seq = tuple((cols, "|S{}".format(dataSize), dataSize) for dataSize, cols in string_cols.items())
        for cols, dtype, dsize in ((integer_cols, numpy.int64, 8),
                                   (double_cols, numpy.double, 8)) + string_seq:

            if len(cols) > 0:

                array = numpy.empty((self.nrows, len(cols)), dtype=dtype, order='C')

                pointer = array.ctypes.data
                strides = array.ctypes.strides

                for i, col in enumerate(cols):

                    lib.odc_decoder_add_column(decoder, col.name.encode('utf-8'))
                    lib.odc_decoder_column_set_data_array(decoder, pos, dsize, strides[0],
                                                          ffi.cast('void*', pointer + (i * strides[1])))
                    pos += 1

                dataframes.append(pandas.DataFrame(array, columns=[c.name for c in cols], copy=False))

        try:
            threads = len(os.shed_getaffinity(0))
        except AttributeError:
            threads = os.cpu_count()

        prows_decoded = ffi.new('long*')
        lib.odc_decode_threaded(decoder, self.__frame, prows_decoded, threads)
        assert prows_decoded[0] == self.nrows

        # Update the missing values (n.b., still sorted by type), and decode strings

        for i in range(len(dataframes)):
            df = dataframes[i]
            if df.dtypes[0] == numpy.int64:
                df.mask(df == missing_integer, inplace=True)
            elif df.dtypes[0] == numpy.double:
                df.mask(df == missing_double, inplace=True)
            else:
                # This is a bit yucky, but I haven't found any other way to decode from b'' strings to real ones
                # Also note, result_type added to work around bug in pandas
                # https://github.com/pandas-dev/pandas/issues/34529
                dataframes[i] = df.apply(lambda x: x.astype('object').str.decode('utf-8'), result_type='expand')

        # And construct the DataFrame from the decoded data

        if len(dataframes) == 1:
            return dataframes[0]
        else:
            return pandas.concat(dataframes, copy=False, axis=1)




"""
At this point, we don't worry about being able to encode with cross-endianness.

That may be useful at some point
"""
import struct
from io import SEEK_SET


class Stream:

    byteOrder = None
    floatMarker = None
    doubleMarker = None

    def __init__(self, f):
        self.f = f
        assert self.byteOrder in ['big', 'little']
        assert self.floatMarker in ['>f', '<f']
        assert self.doubleMarker in ['>d', '<d']

    def seek(self, nbytes):
        self.f.seek(nbytes, SEEK_SET)

    def position(self):
        return self.f.tell()

    # Write functionality

    def write(self, b: bytes):
        self.f.write(b)

    def encodeMarker(self, marker: int):
        self.write(marker.to_bytes(2, byteorder='big', signed=False))

    def encodeUInt8(self, value: int):
        self.write(value.to_bytes(1, byteorder=self.byteOrder, signed=False))

    def encodeUInt16(self, value: int):
        self.write(value.to_bytes(2, byteorder=self.byteOrder, signed=False))

    def encodeInt32(self, value: int):
        self.write(value.to_bytes(4, byteorder=self.byteOrder, signed=True))

    def encodeInt64(self, value: int):
        self.write(value.to_bytes(8, byteorder=self.byteOrder, signed=True))

    def encodeString(self, value: str):
        self.encodeByteString(value.encode('utf-8'))

    def encodeByteString(self, value: bytes):
        self.encodeInt32(len(value))
        self.write(value)

    def encodeReal32(self, value: float):
        self.write(struct.pack(self.floatMarker, value))

    def encodeReal64(self, value: float):
        self.write(struct.pack(self.doubleMarker, value))

    # Read functionality

    def read(self, n: int):
        return self.f.read(n)

    def readMarker(self):
        return int.from_bytes(self.read(2), byteorder='big', signed=False)

    def readUInt8(self):
        return int.from_bytes(self.read(1), byteorder=self.byteOrder, signed=False)

    def readUInt16(self):
        return int.from_bytes(self.read(2), byteorder=self.byteOrder, signed=False)

    def readInt32(self):
        return int.from_bytes(self.read(4), byteorder=self.byteOrder, signed=True)

    def readInt64(self):
        return int.from_bytes(self.read(8), byteorder=self.byteOrder, signed=True)

    def readString(self):
        return self.readByteString().decode('utf-8')

    def readByteString(self):
        return self.read(self.readInt32())

    def readReal32(self):
        return struct.unpack(self.floatMarker, self.read(4))[0]

    def readReal64(self):
        return struct.unpack(self.doubleMarker, self.read(8))[0]


class LittleEndianStream(Stream):

    byteOrder = 'little'
    floatMarker = '<f'
    doubleMarker = '<d'


class BigEndianStream(Stream):

    byteOrder = 'big'
    floatMarker = '>f'
    doubleMarker = '>d'




from .lib import lib, ffi
from .frame import Frame

import pandas
import io

class Reader:
    """This is the main container class for reading ODBs"""

    __reader = None
    __frames = None

    def __init__(self, source, aggregated=True, max_aggregated=-1):

        self.__aggregated = aggregated
        self.__max_aggregated = max_aggregated

        reader = ffi.new('odc_reader_t**')
        if isinstance(source, io.IOBase):
            lib.odc_open_file_descriptor(reader, source.fileno())
        else:
            assert isinstance(source, str)
            lib.odc_open_path(reader, source.encode())

        # Set free function
        self.__reader = ffi.gc(reader[0], lib.odc_close)

    @property
    def frames(self):
        if self.__frames is None:
            self.__frames = []

            frame = ffi.new('odc_frame_t**')
            lib.odc_new_frame(frame, self.__reader)
            frame = ffi.gc(frame[0], lib.odc_free_frame)
            while (lib.odc_next_frame_aggregated(frame, self.__max_aggregated) if self.__aggregated else lib.odc_next_frame(frame)) != lib.ODC_ITERATION_COMPLETE:

                copy_frame = ffi.new('odc_frame_t**')
                lib.odc_copy_frame(frame, copy_frame)
                self.__frames.append(Frame(ffi.gc(copy_frame[0], lib.odc_free_frame)))

        return self.__frames


def _read_odb_generator(source, columns=None, aggregated=True, max_aggregated=-1):
    r = Reader(source, aggregated=aggregated, max_aggregated=max_aggregated)
    for f in r.frames:
        yield f.dataframe(columns)


def _read_odb_oneshot(source, columns=None):
    reduced = pandas.concat(_read_odb_generator(source, columns), sort=False, ignore_index=True)
    for name, data in reduced.iteritems():
        if data.dtype == 'object':
            data.where(pandas.notnull(data), None, inplace=True)
    return reduced


def read_odb(source, columns=None, aggregated=True, single=False, max_aggregated=-1):
    if single:
        assert aggregated
        return _read_odb_oneshot(source, columns)
    else:
        return _read_odb_generator(source, columns, aggregated, max_aggregated)




from .frame import Frame, MismatchedFramesError

import pandas
import io


class Reader:

    def __init__(self, source, aggregated=True):
        self.__aggregated = aggregated
        self._frames = []

        if isinstance(source, io.IOBase):
            self._f = source
        else:
            self._f = open(source, 'rb')

        while True:
            try:
                self._frames.append(Frame(self._f))
                self._frames[-1].seekToEnd()
            except EOFError:
                # n.b. f.read() does not throw EOFError, so this will only catch the exception
                # thrown internally when the table marker is not found
                break

        if self.__aggregated:
            if len(self._frames) > 1:
                aggregated_frames = [self._frames[0]]
                for frame in self._frames[1:]:
                    try:
                        aggregated_frames[-1]._append(frame)
                    except MismatchedFramesError:
                        aggregated_frames.append(frame)
                self._frames = aggregated_frames


    @property
    def frames(self):
        return self._frames


def _read_odb_generator(source, columns=None, aggregated=True):
    r = Reader(source, aggregated=aggregated)
    for f in r.frames:
        yield f.dataframe(columns)


def _read_odb_oneshot(source, columns=None):
    reduced = pandas.concat(_read_odb_generator(source, columns), sort=False, ignore_index=True)
    for name, data in reduced.iteritems():
        if data.dtype == 'object':
            data.where(pandas.notnull(data), None, inplace=True)
    return reduced


def read_odb(source, columns=None, aggregated=True, single=False):
    if single:
        assert aggregated
        return _read_odb_oneshot(source, columns)
    else:
        return _read_odb_generator(source, columns, aggregated)




import io
import os
import re

import setuptools


def read(path):
    file_path = os.path.join(os.path.dirname(__file__), *path.split('/'))
    return io.open(file_path, encoding='utf-8').read()


# single-sourcing the package version using method 1 of:
#   https://packaging.python.org/guides/single-sourcing-package-version/
def parse_version_from(path):
    version_file = read(path)
    version_match = re.search(r"^__version__ = '(.*)'", version_file, re.M)
    if version_match is None or len(version_match.groups()) > 1:
        raise ValueError("couldn't parse version")
    return version_match.group(1)


setuptools.setup(
    name='pyodc',
    version=parse_version_from('codc/__init__.py'),
    description='Python interface to odc for encoding/decoding ODB files',
    long_description=read('README.rst') + read('CHANGELOG.rst'),
    author='European Centre for Medium-Range Weather Forecasts (ECMWF)',
    author_email='software.support@ecmwf.int',
    license='Apache License Version 2.0',
    url='https://github.com/ecmwf/pyodc',
    packages=['pyodc', 'codc'],
    package_data={ '': ['*.h'] },
    include_package_data=True,
    setup_requires=[
        'pytest-runner',
    ],
    install_requires=[
        'cffi',
    ],
    extras_require={},
    tests_require=[
        'pytest',
        'pytest-cov',
        'pytest-flakes',
    ],
    test_suite='tests',
    zip_safe=True,
    keywords='odc odb',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
    ]
)
